from matplotlib.figure import Figure
from matplotlib.axes import Axes
from matplotlib.axes import SubplotBase
import numpy as np

__all__ = [
    "listify_dict",
    "exec_no_show",
    "extract_by_name",
]


def exec_no_show(filepath, globals=None, locals=None):
    """
    exec a file except for plt.show
    """
    if globals is None:
        globals = {}
    globals.update(
        {
            "__file__": filepath,
            "__name__": "__main__",
        }
    )
    with open(filepath) as file:
        block = file.read()
        block = block.replace("plt.show", "")
        exec(compile(block, filepath, "exec"), globals)
    return globals


def listify_dict(d):
    """
    convert a dictionary to a list of names and values filtering only for objects
    that we care about
    """
    names = []
    objs = []
    for k, v in d.items():
        if isinstance(v, (list, tuple, dict, np.ndarray, Axes, SubplotBase, Figure)):
            if isinstance(v, np.ndarray) and not v.dtype == np.object:
                continue
            if "__" in k:
                continue
            names.append(k)
            objs.append(v)
    return names, objs


def extract_by_name(name, globals):
    """
    https://stackoverflow.com/a/63331128/835607
    """
    loc = {}
    exec(f"__fig = {name}", globals, loc)
    return loc["__fig"]



"""
garbo
"""
from mpl_playback.record import record_file

# record_file("dynamic.py", "fig")
# record_file("subplots.py", "ctrls.fig")
# record_file("multifig.py", ["fig", "slider_fig"])
record_file("multifig.py", ["slider_fig", "fig"])
# record_file("multifig.py", ["slider_fig"])



"""
file heavily based on/taken from the sphinx-gallery version
https://github.com/sphinx-gallery/sphinx-gallery/blob/ecd399e2e60557875d9312a6f5f8dbe4d0dd7a0e/sphinx_gallery/scrapers.py
"""
import os
import re
import sys
from distutils.version import LooseVersion
from pathlib import Path
from textwrap import indent
from warnings import filterwarnings

from sphinx.errors import ExtensionError
from sphinx_gallery.utils import optipng, scale_image
from sphinx_gallery.scrapers import matplotlib_scraper as _matplotlib_scraper
from sphinx_gallery.scrapers import figure_rst

from .playback import playback_events
from .record import record_events


def _import_matplotlib():
    """Import matplotlib safely."""
    # make sure that the Agg backend is set before importing any
    # matplotlib
    import matplotlib

    matplotlib.use("agg")
    matplotlib_backend = matplotlib.get_backend().lower()

    filterwarnings(
        "ignore",
        category=UserWarning,
        message="Matplotlib is currently using agg, which is a"
        " non-GUI backend, so cannot show the figure.",
    )

    if matplotlib_backend != "agg":
        raise ExtensionError(
            "Sphinx-Gallery relies on the matplotlib 'agg' backend to "
            "render figures and write them to files. You are "
            "currently using the {} backend. Sphinx-Gallery will "
            "terminate the build now, because changing backends is "
            "not well supported by matplotlib. We advise you to move "
            "sphinx_gallery imports before any matplotlib-dependent "
            "import. Moving sphinx_gallery imports at the top of "
            "your conf.py file should fix this issue".format(matplotlib_backend)
        )

    import matplotlib.pyplot as plt

    return matplotlib, plt


from pathlib import Path

from mpl_playback.playback import load_events, playback_events

SINGLE_IMAGE = """
.. image:: /{}
    :alt: {}
    :class: sphx-glr-single-img
"""


def matplotlib_scraper(block, block_vars, gallery_conf, **kwargs):
    """
    A drop in replacement for the sphinx-gallery matplotlib scraper that will
    also check if there are playback files associated with the example. In that case
    it will generate a gif using mpl-playback. Looks a file ``_<file name>-playback.json``.

    Rest of docstring is taken directly from sphinx-gallery:

    Scrape Matplotlib images.

    Based on the matplotlib_scraper included in sphinx-gallery

    Parameters
    ----------
    block : tuple
        A tuple containing the (label, content, line_number) of the block.
    block_vars : dict
        Dict of block variables.
    gallery_conf : dict
        Contains the configuration of Sphinx-Gallery
    **kwargs : dict
        Additional keyword arguments to pass to
        :meth:`~matplotlib.figure.Figure.savefig`, e.g. ``format='svg'``.
        The ``format`` kwarg in particular is used to set the file extension
        of the output file (currently only 'png', 'jpg', and 'svg' are
        supported).

    Returns
    -------
    rst : str
        The ReSTructuredText that will be rendered to HTML containing
        the images. This is often produced by :func:`figure_rst`.
    """
    # check if there is a playback file
    p = Path(block_vars["src_file"])
    events_path = p.parent.joinpath(f"_{p.stem}-playback.json")

    if not events_path.exists():
        return _matplotlib_scraper(block, block_vars, gallery_conf, **kwargs)

    matplotlib, plt = _import_matplotlib()

    meta, events = load_events(str(events_path))
    figures = meta["figures"]
    outputs = []
    for fig in figures:
        outputs.append(next(block_vars["image_path_iterator"]).replace(".png", ".gif"))
    playback_events(figures, events, meta, block_vars["example_globals"], outputs)

    # print(figure_rst(outputs, gallery_conf['src_dir'], ''))
    # TODO: Allow this to be vertically oriented
    return figure_rst(outputs, gallery_conf["src_dir"], "")



import json
import os
from unittest import mock

import matplotlib

matplotlib.use("Agg")

import numpy as np
from matplotlib import animation
from matplotlib.animation import FFMpegWriter
from .util import exec_no_show, listify_dict, extract_by_name
from ._version import schema_version

_prog_bar = True
try:
    from tqdm import tqdm
except ImportError:
    _prog_bar = False

__all__ = [
    "gen_mock_events",
    "load_events",
    "playback_file",
    "playback_events",
]


def _grab_obj(globals, key):
    if isinstance(key, str):
        return globals.get(key, None)
    elif key is None:
        return None
    else:
        base = globals.get(key[0], None)
        if base is None:
            return base
        # TODO - do this in a clever way that generalizes to n deep nesting
        if len(key) == 2:
            return base[key[1]]
        elif len(key) == 3:
            return base[key[1]][key[2]]
        raise ValueError("Nesting beyond 2 levels not yet supported")


def gen_mock_events(events, globals, accessors):
    times = []
    mock_events = []
    for event in events:
        mock_event = mock.Mock()
        for k, v in event.items():
            if k == "time":
                times.append(v)
            elif k == "fig":
                setattr(mock_event, "canvas", accessors[v].canvas)
                setattr(mock_event, "_figname", v)
            elif k == "inaxes":
                setattr(mock_event, "inaxes", _grab_obj(globals, v))
            else:
                setattr(mock_event, k, v)
        mock_events.append(mock_event)
    return np.array(times), np.array(mock_events)


def load_events(events):
    meta = {}
    if isinstance(events, str):
        with open(events) as f:
            loaded = json.load(f)
            meta["figures"] = loaded["figures"]
            meta["schema-version"] = loaded["schema-version"]
            events = loaded["events"]
    return meta, events


def playback_file(
    events, path, outputs, fps=24, from_first_event=True, prog_bar=True, **kwargs
):
    """
    Parameters
    ----------
    events : str
        Path to the json file defining the events or a dictionary of an
        already loaded file.
    path : str
        path to the file to be executed.
    outputs : str, list of str, or None
        The path(s) to the output file(s). If None then the
        events will played back but no outputs will saved.
    fps : int, default: 24
        Frames per second of the output
    from_first_event : bool, default: True
        Whether playback should start from the timing of the first recorded
        event or from when recording was initiated.
    prog_bar : bool, default: True
        Whether to display a progress bar. If tqdm is not
        available then this kwarg has no effect.
    **kwargs :
        Passed through to `FuncAnimation`.
    """
    if isinstance(outputs, str):
        outputs = [outputs]
    meta, events = load_events(events)
    figures = meta["figures"]
    gbl = exec_no_show(path)
    playback_events(
        figures,
        events,
        meta,
        gbl,
        outputs,
        fps,
        from_first_event,
        prog_bar=prog_bar,
        **kwargs
    )


def playback_events(
    figures,
    events,
    meta,
    globals,
    outputs,
    fps=24,
    from_first_event=True,
    prog_bar=True,
    **kwargs
):
    """
    plays back events that have been

    Parameters
    ----------
    ....
    outputs : list of str or None
        The paths to outputs. If None then the events will played back but
        no outputs will saved.
    fps : int, default: 24
        Frames per second of the output
    from_first_event : bool, default: True
        Whether playback should start from the timing of the first recorded
        event or from when recording was initiated.
    prog_bar : bool, default: True
        Whether to display a progress bar. If tqdm is not
        available then this kwarg has no effect.
    """
    accessors = {}
    fake_cursors = {}
    writers = []
    _figs = {}  # the actual figure objects
    transforms = {}
    for fig, out in zip(figures, outputs):
        _fig = extract_by_name(fig, globals)
        _figs[fig] = _fig
        # pre invert for performance (i have no idea if this owuld get cached.)
        # but if it's inverting a matrix then it's bad news to constantly invert
        transforms[fig] = _fig.transFigure.inverted().frozen()
        accessors[fig] = _fig
        fake_cursors[fig] = _fig.axes[-1].plot(
            [0, 0],
            [0, 0],
            "k",
            marker=6,
            markersize=15,
            transform=_fig.transFigure,  # confusing that this is necessary, see note in animate below
            clip_on=False,
            zorder=99999,
        )[0]
        if outputs is not None:
            writers.append(FFMpegWriter(fps, out))
            writers[-1].setup(_fig, out, len(events))

    times, mock_events = gen_mock_events(events, globals, accessors)
    if from_first_event:
        times -= times[0]
    N_frames = np.int(times[-1] * fps)
    # map from frames to events
    event_frames = np.round(times * fps)

    if prog_bar and _prog_bar:
        pbar = tqdm(total=N_frames)

    def animate(i):
        idx = event_frames == i
        if np.sum(idx) != 0:
            for event in mock_events[idx]:
                # event = mock_events[i]
                accessors[event._figname].canvas.callbacks.process(event.name, event)
                if event.name == "motion_notify_event":
                    # now set the cursor invisible so multiple don't show up
                    # if there are multiple figures
                    for fc in fake_cursors.values():
                        fc.set_visible(False)

                    # It really seems as though this transform should be uncessary
                    # and with only a single figure it is uncesssary. But for reasons that
                    # are beyond me, when there are multiple figures this transform is crucial
                    # or else the cursor will show up in a weirdly scaled position
                    # this is true even with setting `transform=None` on the origin `plot` call
                    f = _figs[event._figname]
                    xy = transforms[event._figname].transform([event.x, event.y])
                    fake_cursors[event._figname].set_data([xy[0]], [xy[1]])
                    fake_cursors[event._figname].set_visible(True)

                # theres got to be a clever way to avoid doing these gazillion draws
                # maybe monkeypatching the figure's draw event?
                for f in _figs.values():
                    f.canvas.draw()

        if prog_bar and _prog_bar:
            pbar.update(1)
        if outputs is not None:
            for w in writers:
                w.grab_frame()

    for i in range(N_frames):
        animate(i)

    if outputs is not None:
        for w in writers:
            w.finish()



"""
==========================
Also supports static files
==========================

This file is the same as :doc:`/gallery/dynamic` but
does not have an associated playback file, so the scraper falls back to
the sphinx-gallery matplotlib scraper.
"""
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.widgets import Button, Slider


# The parametrized function to be plotted
def f(t, amplitude, frequency):
    return amplitude * np.sin(2 * np.pi * frequency * t)


t = np.arange(0.0, 1.0, 0.001)

# Define initial parameters
init_amplitude = 5
init_frequency = 3

# Create the figure and the line that we will manipulate
fig, ax = plt.subplots()
(line,) = plt.plot(t, f(t, init_amplitude, init_frequency), lw=2)

axcolor = "lightgoldenrodyellow"
ax.margins(x=0)

# adjust the main plot to make room for the sliders
plt.subplots_adjust(left=0.25, bottom=0.25)

# Make a horizontal slider to control the frequency.
axfreq = plt.axes([0.25, 0.1, 0.65, 0.03], facecolor=axcolor)
freq_slider = Slider(
    ax=axfreq,
    label="Frequency",
    valmin=0.1,
    valmax=30.0,
    valinit=init_frequency,
)

# Make a vertically oriented slider to control the amplitude
axamp = plt.axes([0.1, 0.25, 0.0225, 0.63], facecolor=axcolor)
amp_slider = Slider(
    ax=axamp,
    label="Amplitude",
    valmin=0.1,
    valmax=10.0,
    valinit=init_amplitude,
    orientation="vertical",
)


# The function to be called anytime a slider's value changes
def update(val):
    line.set_ydata(f(t, amp_slider.val, freq_slider.val))
    fig.canvas.draw_idle()


# register the update function with each slider
freq_slider.on_changed(update)
amp_slider.on_changed(update)

# Create a `matplotlib.widgets.Button` to reset the sliders to initial values.
resetax = plt.axes([0.8, 0.025, 0.1, 0.04])
button = Button(resetax, "Reset", color=axcolor, hovercolor="0.975")


def reset(event):
    freq_slider.reset()
    amp_slider.reset()


button.on_clicked(reset)

plt.show()



import collections
import json
from functools import partial
from os import path
from pathlib import Path

from .util import exec_no_show, listify_dict, extract_by_name
from ._version import schema_version
from time import time

__all__ = [
    "possible_events",
    "record_events",
    "record_file",
    "record_figure",
]
possible_events = [
    "button_press_event",
    "button_release_event",
    "draw_event",
    "key_press_event",
    "key_release_event",
    "motion_notify_event",
    "pick_event",
    "resize_event",
    "scroll_event",
    "figure_enter_event",
    "figure_leave_event",
    "axes_enter_event",
    "axes_leave_event",
]

event_list = []


from matplotlib import axes


import numpy as np


def _find_obj(names, objs, obj, accessors):
    """
    accessors : dict
        A dict of predefined ways to access objects.

    find the name of the figure or axes.
    TODO: Checks in nested lists and tuples
    TODO: check in dictionaries
    """
    if obj is None:
        return None
    if obj in accessors:
        return accessors[obj]

    for name, maybe in zip(names, objs):
        if obj is maybe:
            return name
        if isinstance(maybe, np.ndarray):
            # gotta special case otherwise potential
            # for ValueErrors when doing obj in numpy-array
            if maybe.dtype == np.object:
                if obj in maybe:
                    return name, int(np.where(maybe == obj)[0][0])
        elif isinstance(maybe, (list, tuple)) and obj in maybe:
            return name, maybe.index(obj)
        # elif isinstance(maybe, dict) and obj in maybe:
        #     return _find_in_dict(maybe, dict)
    raise ValueError(
        f"Something has gone wrong while encoding: {str(obj)}"
        " - please report this issue on https://github.com/ianhi/mpl-playback"
    )


def _record_event(event, fig, names, objs, accessors, start_time):
    info2save = [
        e for e in dir(event) if "_" not in e and e not in ["guiEvent", "lastevent"]
    ]
    saved_info = {k: getattr(event, k) for k in info2save}
    saved_info.pop("canvas")
    saved_info["fig"] = _find_obj(names, objs, event.canvas.figure, accessors)
    saved_info["inaxes"] = _find_obj(names, objs, saved_info["inaxes"], accessors)
    saved_info["time"] = time() - start_time
    event_list.append(saved_info)


def record_file(path, fig="fig", output=None):
    """
    Parameters
    ----------

    path : str
        The path to the file to be run
    fig : str, default: fig
        The variable name of the figure to capture. Can
        also be a string that access objects. e.g.:
        ``controls.fig`` or ``list_of_figures[0][1]``.
    output : str or None, default: None
        Defaults to ``_<file-name>-playback.json``
    """
    globals = exec_no_show(path)

    if isinstance(fig, str):
        fig = [fig]
    if output is None:
        output = "_" + Path(path).stem + "-playback.json"
    record_figures(fig, globals, output)


def record_figure(fig, globals, savename, accessors=None):
    record_figures([fig], globals, savename, accessors=None)


def record_figures(figures, globals, savename, accessors=None):
    """
    Parameters
    ----------
    fig : list of str
        list of exec-able calls to obtain the figure.
    globals : dict
    savename : str
    accessors : dict
        Ways to access objects. Useful for figures or axes
    """
    if accessors is None:
        accessors = {}

    for fig in figures:
        _fig = extract_by_name(fig, globals)
        accessors[_fig] = fig
        record_events(
            _fig,
            ["motion_notify_event", "button_press_event", "button_release_event"],
            globals,
            accessors,
        )
    globals["plt"].show()
    with open(savename, "w") as fp:
        json.dump(
            {
                "figures": figures,
                "schema-version": schema_version,
                "events": event_list,
            },
            fp,
        )


def record_events(fig, events, globals, accessors=None):
    start_time = time()
    if accessors is None:
        accessors = {}
    names, objs = listify_dict(globals)
    if isinstance(events, str):
        events = [events]
    for e in events:
        fig.canvas.mpl_connect(
            e,
            partial(
                _record_event,
                fig=fig,
                names=names,
                objs=objs,
                accessors=accessors,
                start_time=start_time,
            ),
        )



"""
================
Multiple Figures
================

Test case for putting multiple figures into sphinx gallery.

The playback file was generated using:
``record_file("multifig.py", ["slider_fig", "fig"])``

"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider

# The parametrized function to be plotted
def f(t, amplitude, frequency):
    return amplitude * np.sin(2 * np.pi * frequency * t)


t = np.linspace(0, 1, 1000)

# Define initial parameters
init_amplitude = 5
init_frequency = 3

# Create the figure and the line that we will manipulate
fig, ax = plt.subplots()
(line,) = plt.plot(t, f(t, init_amplitude, init_frequency), lw=2)


# The function to be called anytime a slider's value changes
def update(val):
    line.set_ydata(f(t, 5, val))
    fig.canvas.draw_idle()


slider_fig, s_ax = plt.subplots(figsize=(6.4, 2))
slider = Slider(s_ax, "freq", 0, 30, valinit=3)
slider.on_changed(update)
plt.show()



from mpl_playback.playback import playback_file

# playback_file("_dynamic-playback.json", "dynamic.py", "ohboi.gif")
# playback_file("_subplots-playback.json", "subplots.py", "ohboi.gif")
playback_file("_multifig-playback.json", "multifig.py", ["multi1.gif", "multi2.gif"])
# playback_file("_multifig-playback.json", "multifig.py", ["multi1-garbo.gif", "multi2-garbo.gif"])

# playback_file("_multifig-playback.json", "multifig.py", ["multi.gif"])
